package org.kozyrev.gui.barchart;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.Line2D;

import javax.swing.BoundedRangeModel;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.JComponent;

import org.kozyrev.gui.barchart.BarChartModel.ModelChangeListener;

public class BarChart extends JComponent {
    
    private static final long serialVersionUID = 6062677098610626789L;

    private final int BORDER_WIDTH = 7;
    private final int RULLER_COUNT = 6;
    private final int BAR_WIDTH = 18;
    private final int STICK_SIZE = 4;
    
    private BarChartModel myModel = new DefaultBarChartModel();
    private ModelChangeListener myListener;

    private int scaleSize = 80;
    private int barFieldSize = scaleSize*RULLER_COUNT;
        


    private BoundedRangeModel rangeModel = new DefaultBoundedRangeModel();
   
    public BarChart(BarChartModel model) {
        if (model != null) {
            setBarChartModel(model);    
        }        
        
        myModel.addModelChangeListener(myListener = new ModelChangeListener() {
            public void modelChanged() {
                BarChart.this.revalidate();
                BarChart.this.repaint();
            }            
        });

    }
    
    public BarChart() {
        this(null);
        rangeModel.setMinimum(0);
        rangeModel.setMaximum(0);
        rangeModel.setValue(0);
    }
    
    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        
        Dimension screen = this.getSize();
        
        if (myModel.getBarCount() == 0) {
            g.drawString("   ", 15, 15);
            return;
        }
        
        String[] strings = new String[myModel.getBarCount()];
        for (int i = 0; i < myModel.getBarCount(); i++) {
            strings[i] = myModel.getBarText(i);
        }
        
        int textFieldSize = getMaxTextWidth(g.getFontMetrics(), strings);
        
        scaleSize = (screen.width - textFieldSize - 4*BORDER_WIDTH) / RULLER_COUNT;
        barFieldSize = scaleSize*RULLER_COUNT;
        
        //////draw stiks
        
        for (int i = 1; i <= myModel.getBarCount(); i++) {
            //draw stick
           
            int currentBarVerticalLocation = BORDER_WIDTH + (2 * BORDER_WIDTH + BAR_WIDTH)*i;
            int rullerVerticalLocation = screen.height - 2*BORDER_WIDTH;
            
            if (currentBarVerticalLocation < rullerVerticalLocation) {
                
                int stickX1 = 2*BORDER_WIDTH + textFieldSize - STICK_SIZE;
                int stickY1 = BORDER_WIDTH + (2*BORDER_WIDTH + BAR_WIDTH)*i;
                int stickX2 = 2*BORDER_WIDTH + textFieldSize;
                int stickY2 = stickY1;
                int stickWidth = 1;
                Color stickColor = Color.GRAY;
                
                drawBoldLine(g, stickX1, stickY1, stickX2, stickY2, stickWidth, stickColor);                
            }
           
        }
        
        ////draw rulers
        
        /////finding max bar weight
        
        int maxWeight = 0;
        for (int i = 0; i < myModel.getBarCount(); i++) {
            if (myModel.getBarWeight(i) > maxWeight) {
                maxWeight = myModel.getBarWeight(i); 
            }
        }
        
        double zoomFactor = ((double) barFieldSize) / ((double) maxWeight);
        double pixCost = 1 / zoomFactor;
        
        //scale normalization
        double cost = scaleSize * pixCost;
        
        cost = Math.ceil(cost);
        
        if (cost > 10000) {
            cost = 10000 * Math.ceil(cost / 10000);
        } else {
            if (cost > 1000) {
                cost = 1000 * Math.ceil(cost / 1000);
            } else {
                if (cost > 100) {
                    cost = 100 * Math.ceil(cost / 100);
                } else {
                    if (cost > 10) {
                        cost = 10 * Math.ceil(cost / 10);
                    } else {
                        if (cost > 5) {
                            cost = 5 * Math.ceil(cost / 5);
                        } else {
                            if (cost > 2) {
                                cost = 2 * Math.ceil(cost / 2);
                            }
                        }
                    }    
                }    
            }
        }               
        
        pixCost = cost / scaleSize;
        zoomFactor = 1 / pixCost;
        
        for (int i = 1; i <= RULLER_COUNT; i++) {
            //draw scale
                   
            int scaleHorisLocation = 2*BORDER_WIDTH + textFieldSize + i*scaleSize;
            int scaleY1 = BORDER_WIDTH;
            int scaleY2 = screen.height - 2*BORDER_WIDTH;
            int scaleWidth = 3;
            Color scaleColor = Color.GRAY;
            
            drawMarkerLine(g, scaleHorisLocation, scaleY1, 
                    scaleHorisLocation, scaleY2, 
                    scaleWidth, scaleColor);
            
            //draw text
            
            int textWidth = g.getFontMetrics().stringWidth((new Integer((int)Math.round(i*scaleSize * pixCost))).toString());
            String text = (new Integer((int)Math.round(i*scaleSize * pixCost))).toString();
            int textHorizLocation = 2*BORDER_WIDTH + textFieldSize + i*scaleSize - textWidth / 2;
            
            g.drawString(text,
                    textHorizLocation, screen.height - 1);
        }
        
        /////draw text
        
        for (int i = 0; i < strings.length; i++) {
            int barVertLocation = (2*BORDER_WIDTH + BAR_WIDTH) * (i+1);
            int canvasVertSize = screen.height - 2*BORDER_WIDTH;
            
            if (barVertLocation < canvasVertSize
                    && rangeModel.getValue() + i < strings.length) {
                
                String text = strings[rangeModel.getValue() + i];
                int textHorizLocation = BORDER_WIDTH + textFieldSize - g.getFontMetrics().stringWidth(strings[rangeModel.getValue() + i]);
                int textVertLocation = (2*BORDER_WIDTH + BAR_WIDTH) * (i+1);
                
                g.drawString(text, textHorizLocation, textVertLocation);                
            }
        }
        
        ////draw bars
        
        for (int i = 1; i <= myModel.getBarCount(); i++) {
            int barVertLocation = (2*BORDER_WIDTH + BAR_WIDTH) * i;
            int canvasVertSize = screen.height - 2*BORDER_WIDTH;
            
            if (barVertLocation < canvasVertSize
                    && rangeModel.getValue() + i - 1 < myModel.getBarCount()) {
                
                int barX1 = 2*BORDER_WIDTH + textFieldSize;
                int barY1 = (2*BORDER_WIDTH + BAR_WIDTH) * i - BAR_WIDTH / 2;
                int barX2 = (int) Math.round(2*BORDER_WIDTH + textFieldSize + myModel.getBarWeight(rangeModel.getValue() + i - 1) * zoomFactor); 
                int barY2 = (2*BORDER_WIDTH + BAR_WIDTH) * i - BAR_WIDTH / 2;  
                Color barColor = myModel.getColor(rangeModel.getValue() + i - 1);
                
                drawBar(g, barX1, barY1, 
                        barX2, barY2,
                        BAR_WIDTH, barColor);                
            }            
        }
        
        
        //draw Axes
        ////draw x axis
        drawBoldLine(g, 2*BORDER_WIDTH + textFieldSize, BORDER_WIDTH, 
                2*BORDER_WIDTH + textFieldSize, 
                screen.height - 2*BORDER_WIDTH, 2, Color.GRAY);

        ////draw y axis
        drawBoldLine(g, 2*BORDER_WIDTH + textFieldSize, screen.height - 2*BORDER_WIDTH,
                2*BORDER_WIDTH + textFieldSize + barFieldSize, screen.height - 2*BORDER_WIDTH, 2, Color.GRAY);
        
    }
    
    public void setBarChartModel(BarChartModel model) {
        myModel.removeModelChangeListener(myListener);
        
        myModel = model;
        
        rangeModel.setMinimum(0);
        rangeModel.setMaximum(myModel.getBarCount() - 1);
        
        myModel.addModelChangeListener(myListener = new ModelChangeListener() {
            public void modelChanged() {
                rangeModel.setMinimum(0);
                rangeModel.setMaximum(myModel.getBarCount() - 1);
                BarChart.this.repaint();                  
            }            
        });
    }
    
    public BarChartModel getBarChartModel() {
        return myModel;
    }
    
    
    public BoundedRangeModel getBoundedRangeModel() {
        return rangeModel;
    }

    /**
     * Draws bold lines with given width of marker
     * 
     * @param g Graphics context
     * @param x1 top left corner 
     * @param y1 top left corner
     * @param x2 bottom right corner
     * @param y2 bottom right corner
     * @param width width of marker
     */
    private void drawBoldLine(Graphics g, int x1, int y1, int x2, int y2, int width, Color c) {    
        
        Graphics2D g2 = (Graphics2D) g;
        
        Stroke stroke = g2.getStroke();
        Paint paint = g2.getPaint();
        
        g2.setPaint(c);
        g2.setStroke(new BasicStroke((float) width));
        g2.draw(new Line2D.Double(x1, y1, x2, y2));
        
        g2.setStroke(stroke);
        g2.setPaint(paint);
    }

    private void drawBar(Graphics g, int x1, int y1, int x2, int y2, int width, Color c) {    
        
        Graphics2D g2 = (Graphics2D) g;
        
        Stroke stroke = g2.getStroke();
        Paint paint = g2.getPaint();
        
        try {
            GradientPaint grad = new GradientPaint(x1,y1, Color.DARK_GRAY, x2, y2, c);
            g2.setPaint(grad);    
            
            if (width <= 1) {
                g.drawLine(x1, y1, x2, y2);
                return;
            }
            
            double r = Math.sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2));
           
            int dx = (int) Math.floor(width/2 * Math.abs(y1 - y2) / r);
            int dy = (int) Math.floor(width / 2 * Math.abs(x1 - x2) / r);
            
            g.fillPolygon(new Polygon(new int[] {x1 - dx, x2 - dx, x2 + dx, x1 + dx}, 
                                      new int[] {y1 + dy, y2 + dy, y2 - dy, y1 - dy}, 4));
        } finally {
            g2.setStroke(stroke);
            g2.setPaint(paint);
        }
    }

    /**
     * Draws marker line
     * 
     * @param g graphics contexts
     * @param x1
     * @param y1
     * @param x2
     * @param y2
     * @param miterlimit dash size 
     */
    private void drawMarkerLine(Graphics g, int x1, int y1, int x2, int y2, int miterlimit, Color c) {
        Graphics2D g2 = (Graphics2D) g;
    
        Stroke stroke = g2.getStroke();
        Paint paint = g2.getPaint();
        
        g2.setPaint(c);
        
        float dash[] = {(float) miterlimit};
        g2.setStroke(new BasicStroke(1.0f,
            BasicStroke.CAP_BUTT,
            BasicStroke.JOIN_MITER,
            (float) miterlimit, dash, 0.0f));
        
        g2.draw(new Line2D.Double(x1, y1, x2, y2));
    
        g2.setStroke(stroke);
        g2.setPaint(paint);
    }

    private int getMaxTextWidth(FontMetrics metrics, String[] strings) {
        
        int max = 0;
        for (int i = 0; i < strings.length; i++) {
            if (metrics.stringWidth(strings[i]) > max) {
                max = metrics.stringWidth(strings[i]);
            }
        }
        
        return max;
    }
    
}
