package org.kozyrev.model.xml;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.kozyrev.model.ClassRecord;
import org.kozyrev.model.FrameRecord;
import org.kozyrev.model.MethodRecord;
import org.kozyrev.model.ProfileModel;
import org.kozyrev.model.ThreadRecord;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class DomXMLParser {

    private static DocumentBuilder documentBuilder;
    private static Document doc;
    private static HashMap<String, ClassRecord> classMap;
    
    private final static int ELEMENT_TYPE = 1;
    
    private final static String ATTRIBUTE_CLASS_NAME = "cn";
    private final static String ATTRIBUTE_METHOD_NAME = "mn";
    private final static String ATTRIBUTE_COUNT = "c";
    private final static String ATTRIBUTE_SHORT_NAME = "s";
    private final static String ATTRIBUTE_FULL_NAME = "f";
    private final static String ATTRIBUTE_ID = "id";
    private final static String ATTRIBUTE_NANOTIME = "t";
        
    
    
    //  The list of elements to display in the tree
    private static String[] treeElementNames = {
        "profile",
        "thread",
        "interaction",         
        "frame", 
        "class-map",
        "entry",
        "allocation",
        "class" 
    };

    public static ProfileModel parseFile(String fileName) {
        File file = new File(fileName);        
        return parseFile(file);
    }
    
    public static ProfileModel parseFile(File file) {
        
        if (documentBuilder == null) {
            try {
                documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            } catch (ParserConfigurationException e) {
                e.printStackTrace();
            } 
        }
        
        try {
            doc = documentBuilder.parse(file);
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        
        Node peek = doc.getFirstChild(); //<profile>
        Node allocation = null;
        Node classes = null;
        
        //Preparing frame parsing. Creating ClassRecord for all classes
        NodeList childNodes = peek.getChildNodes();
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node node = childNodes.item(i);
            
            //System.out.println(node.getNodeName() + ":" + node.getNodeType());
            
            if (node.getNodeType() == ELEMENT_TYPE 
                    && node.getNodeName().equals("allocation")) {
                
                allocation = node;
                
            } else {
                if (node.getNodeType() == ELEMENT_TYPE 
                        && node.getNodeName().equals("class-map")) {
                    
                    classes = node;
                } 
            }          
        }
        
        classMap = parseClasses(allocation, classes);
        
        //getting results
        ArrayList<ThreadRecord> threads = new ArrayList<ThreadRecord>();
        
        childNodes = peek.getChildNodes();
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node node = childNodes.item(i);
            
            if (node.getNodeType() == ELEMENT_TYPE 
                    && node.getNodeName().equals("thread")) {
                
                ThreadRecord thread = 
                    new ThreadRecord(node.getAttributes().getNamedItem(ATTRIBUTE_ID).getNodeValue()); 
                
                NodeList interactions = node.getChildNodes();
                
                for (int j = 0; j < interactions.getLength(); j++) {
                    node = interactions.item(j);
                    
                    if (node.getNodeType() == ELEMENT_TYPE 
                            && node.getNodeName().equals("interaction")) {
                        
                        NodeList childs = node.getChildNodes();
                        
                        for (int k = 0; k < childs.getLength(); k++) {
                            
                            node = childs.item(k);
                            if (node.getNodeType() == ELEMENT_TYPE &&
                                    node.getNodeName().equals("frame")) {
                                
                                thread.addFrame(parseFrame(node));    
                            }                           
                        }                        
                    }
                }
            
                threads.add(thread);
            }    
        }
       
        return new ProfileModel(getClasses(), threads);
    }
    
    private static FrameRecord parseFrame(Node node) {
        
        NamedNodeMap attributes = node.getAttributes();
        String classShortName = attributes.getNamedItem(ATTRIBUTE_CLASS_NAME).getNodeValue();
        String methodFullName = attributes.getNamedItem(ATTRIBUTE_METHOD_NAME).getNodeValue();
        int count = Integer.parseInt(attributes.getNamedItem(ATTRIBUTE_COUNT).getNodeValue());
        long nanos = Long.parseLong(attributes.getNamedItem(
                ATTRIBUTE_NANOTIME).getNodeValue());
        
        ClassRecord methodClass = classMap.get(classShortName);
        MethodRecord method = new MethodRecord(methodFullName, methodClass);
        
        FrameRecord child = null;
        FrameRecord result = new FrameRecord(method, count, nanos);
        
        if (node.hasChildNodes()) {
            
            NodeList childs = node.getChildNodes();
            for (int i = 0; i < childs.getLength(); i++) {
                
                Node childNode = childs.item(i);
                if (childNode.getNodeType() == ELEMENT_TYPE
                        && childNode.getNodeName().equals("frame")) {
                    
                    child = parseFrame(childNode);
                    
                    result.addFrame(child);
                }    
            }
            
        }
        
        return result;
    }

    private static HashMap<String, ClassRecord> parseClasses(Node allocation, Node classes) {
        
        HashMap<String, ClassRecord> classMap = new HashMap<String, ClassRecord>();
        
        HashMap<String, String> names = new HashMap<String, String>();
        HashMap<String, Integer> counts = new HashMap<String, Integer>();
        
        //finding alloc. counts
        NodeList childNodes = allocation.getChildNodes();
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node node = childNodes.item(i);
            
            if (node.getNodeType() == ELEMENT_TYPE 
             && treeElement( node.getNodeName())) {
                
                counts.put(node.getAttributes().getNamedItem(ATTRIBUTE_CLASS_NAME).getNodeValue(),
                        Integer.parseInt(node.getAttributes().getNamedItem(ATTRIBUTE_COUNT).getNodeValue()));
                
            }
    
        }
        
        //finding full names
        childNodes = classes.getChildNodes();
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node node = childNodes.item(i);
            
            if (node.getNodeType() == ELEMENT_TYPE 
             && treeElement( node.getNodeName())) {
                
                names.put(node.getAttributes().getNamedItem(ATTRIBUTE_SHORT_NAME).getNodeValue(),
                        node.getAttributes().getNamedItem(ATTRIBUTE_FULL_NAME).getNodeValue());
                
            }
    
        }
        
        //agregating information and constructing map from class name to ClassRecord
        for (String shortName : names.keySet()) {
            
            String fullName = names.get(shortName);
            Integer count = counts.get(shortName);
            
            if (count == null) {
                count = Integer.valueOf(0);
            }
            
            ClassRecord cr = new ClassRecord(fullName, shortName, count.intValue());
            
            classMap.put(shortName, cr);
        }
        
        return classMap;
    }

    private static Boolean treeElement(String elementName) {
      for (int i=0; i<treeElementNames.length; i++) {
        if ( elementName.equals(treeElementNames[i]) ) return true;
      }
      return false;
    }

    private static List<ClassRecord> getClasses() {
        
        ArrayList<ClassRecord> result = new ArrayList<ClassRecord>();
        for (ClassRecord rec : classMap.values()) {
            result.add(rec);
        }
        
        return result;
    }
  
}
